'''
tests - Tests support library for "dns_sprockets".

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''

# end of file



'''
dns_utils - Misc. DNS utility functions not present in dnspython.
-----------------------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import dns.name


def is_delegated(delegated_names, name):
    '''
    Tests a name to see if it is in a delegated zone.

    :param list delegated_names: List of non-apex dns.name.Names in zone with NS RRSets.
    :param obj name: The dns.name.Name to test.
    :return: True if name is delegated, False otherwise.
    '''
    for deleg_name in delegated_names:
        if name.is_subdomain(deleg_name):
            return True
    return False


def dns_name_cmp_to_key():
    '''
    Convert dns.name.Name.fullcompare method into a key= "function", which is a
    class suitable for the key= argument for sort/sorted.

    :return: Class that uses fullcompare() on dns.name.Name's.
    '''
    class Key(object):
        # pylint: disable=too-few-public-methods
        '''
        Wraps a dns.name.Name object with comparison operators that use the
        dns.name.Name.fullcompare() method.
        '''

        def __init__(self, obj, *_args):
            # pylint: disable=unused-argument
            ''' Ctor, initializes a Key instance. '''
            self.obj = obj

        def __lt__(self, other):
            ''' Less-than operator. '''
            return self.obj.fullcompare(other.obj)[1] < 0

        def __gt__(self, other):
            ''' Greater-than operator. '''
            return self.obj.fullcompare(other.obj)[1] > 0

        def __eq__(self, other):
            ''' Equality operator. '''
            return self.obj.fullcompare(other.obj)[1] == 0

        def __le__(self, other):
            ''' Less-than-or-equal operator. '''
            return self.obj.fullcompare(other.obj)[1] <= 0

        def __ge__(self, other):
            ''' Greater-than-or-equal operator. '''
            return self.obj.fullcompare(other.obj)[1] >= 0

        def __ne__(self, other):
            ''' Non-equality operator. '''
            return self.obj.fullcompare(other.obj)[1] != 0

    return Key


def dnssec_sort_names(names, reverse=False):
    '''
    Sorts names to DNSSEC sorted order.

    :param names: Container of dns.name.Names.
    :return: List of DNSSEC-sorted order dns.name.Names.
    '''
    return sorted(names, key=dns_name_cmp_to_key(), reverse=reverse)


def calc_node_names(node_names, ents_too, zone_name):
    '''
    Calculates list of node names in a zone (including any Empty Non-Terminals
    if 'ents_too' is True), ordered in DNSSEC name order.

    :param list node_names: The list of name objects from Zone.nodes.keys().
    :param bool ents_too: Set True to generate ENT names too.
    :param obj zone_name: Name of zone (used when 'ents_too' is True).
    :return: Sorted list of all node names.
    '''

    # Generate set of names (for lookups):
    names = frozenset(node_names)

    # If we want ENT's too, spin through names (in DNSSEC-order) and add them:
    ent_names = set()
    if ents_too:
        for name in dnssec_sort_names(node_names):

            # Chop-off most specific label and see if it is accounted for:
            pot_ent_name = dns.name.Name(name.labels[1:])
            while pot_ent_name.is_subdomain(zone_name):

                if pot_ent_name not in names:
                    ent_names.add(pot_ent_name)

                # Chop off next most specific lable and continue:
                pot_ent_name = dns.name.Name(pot_ent_name.labels[1:])

    # Gather all names and sort in DNSEC name order:
    return dnssec_sort_names(names | ent_names)

# end of file



'''
rrsig_orphan - Record test: RrsigOrphan

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import time

import dns.rdtypes.ANY.RRSIG
import dns.dnssec

import dns_sprockets_lib.validators as validators


class RrsigOrphan(validators.RecTest):
    # pylint: disable=too-few-public-methods
    '''
    Checks for orphan RRSIGs.
    '''
    TEST_DNSSECTYPE = True
    TEST_RRTYPE = 'RRSIG'
    TEST_OPTARGS = {
        'now': (None, 'Time to use for validating RRSIG time windows, e.g. 20150101123000'),
        'now_offset': (None, 'Number of seconds to offset the "now" value, e.g. -86400)')}

    def __init__(self, args):

        self.now = None
        self.now_offset = None
        super(RrsigOrphan, self).__init__(args)

        self.posix_now = (self.now
            and dns.rdtypes.ANY.RRSIG.sigtime_to_posixtime(self.now)
            or int(time.time()))
        if self.now_offset:
            self.posix_now += int(self.now_offset)

    def run(self, context, suggested_tested, name, ttl, rdata):
        # pylint: disable=too-many-arguments

        result = None

        # Make sure there's a covered RRSet for the RRSIG rdata:
        rdataset = context.zone_obj.get_rdataset(name, rdata.type_covered)
        if not rdataset:
            result = 'No RRSet for name: %s type: %s' % (
                name, dns.rdatatype.to_text(rdata.type_covered))
        else:
            try:
                dns.dnssec.validate_rrsig(
                    (name, rdataset),
                    rdata,
                    {context.zone_name: context.dnskey_rdataset},
                    now=self.posix_now)
            except dns.dnssec.UnsupportedAlgorithm as err:
                result = str(err)
            except dns.dnssec.ValidationFailure as err:
                result = str(err)

        return (suggested_tested, result)

# end of file



#!/usr/bin/env python2.7
'''
dns_sprockets - A command-line tool to validate DNS zones.
----------------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import os
import sys
import time

from dns_sprockets_version import VERSION
import dns_sprockets_lib.utils as utils
import dns_sprockets_lib.loaders as loaders
import dns_sprockets_lib.loader_classes as loader_classes
import dns_sprockets_lib.validators as validators
import dns_sprockets_lib.validator_classes as validator_classes
from dns_sprockets_lib.dns_sprockets_impl import DNSSprocketsImpl


_PRODUCT = 'dns_sprockets'
_BANNER = '%s (%s) - A DNS Zone validation tool' % (_PRODUCT, VERSION)
_DEFAULT_ZONE = 'www.ultradns.com'
_DEFAULT_LOADER = 'xfr'
_DEFAULT_SOURCE = '127.0.0.1#53'
_FATAL_RETCODE = 255


def _run_main(avail_loaders, avail_tests, args):
    '''
    Constructs and runs a DNSSprocketsImpl instance.
    '''
    start_time = time.time()

    # Construct and run a DNSSprocketsImpl instance:
    try:
        (load_time, _, ret_code) = DNSSprocketsImpl(avail_loaders, avail_tests, args).run()
        total_time = time.time() - start_time
        print('# TOTAL ELAPSED TIME: %f SECS  LOAD TIME: %f SECS  TEST TIME: %f SECS' % (
            total_time, load_time, total_time - load_time))
        sys.stdout.flush()
        os._exit(ret_code >= _FATAL_RETCODE and _FATAL_RETCODE - 1 or ret_code)

    except Exception as err:
        print('FATAL: {%s} %s' % (err.__class__.__name__, err))
        if args.verbose:
            raise
        sys.stdout.flush()
        os._exit(_FATAL_RETCODE)


def run():
    '''
    Run the sprocket command-line application.
    '''
    # Print startup banner:
    print('#', _BANNER)

    # Determine available loaders and tests:
    avail_loaders = utils.public_modules_in_package(loaders, ['tests'])
    avail_tests = utils.public_modules_in_package(validators, ['tests'])

    # Parse command-line:
    import argparse
    parser = argparse.ArgumentParser(
        epilog='''
Use @filename to read some/all arguments from a file.\n\n
Use -d's to define optional, module-specific parameters if desired (e.g. to tell
'xfr' loader to use a specific source address, use "-d xfr_source=1.2.3.4").
The optional parameters are listed under each loader and validator description
in DEFINE lines, if available.  The "short form" of parameter names (i.e. 
without the module name prefix) may be used instead, if multiple modules define
the same "short form" name (e.g. -d now=20160328120000 can be used, and will set
both 'rrsig_missing' and 'rrsig_orphan' validators' rrsig_missing_now and 
rrsig_orphan_now values, respectively).\n\n
By default, all tests are run.  Use -i\'s to explicitly specify desired tests,
or -x\'s to eliminate undesired tests.\n\n
The list of available loaders is:
---------------------------------------------------------------------------\n%s\n\n
The list of available validators is:
---------------------------------------------------------------------------\n%s
            ''' % (loader_classes.get_formatted_descriptions(),
                   validator_classes.get_formatted_descriptions()),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        fromfile_prefix_chars='@')
    parser.add_argument('-z', '--zone', dest='zone',
        default=_DEFAULT_ZONE, metavar='s',
        help='Name of zone to validate [%(default)s]')
    parser.add_argument('-l', '--loader', dest='loader',
        default=_DEFAULT_LOADER, metavar='s', choices=avail_loaders,
        help='Zone loader method to use (one of: %(choices)s) [%(default)s]')
    parser.add_argument('-s', '--source', dest='source',
        default=_DEFAULT_SOURCE, metavar='s',
        help='Loader source to use [%(default)s]')
    parser.add_argument('-i', '--include', dest='include_tests',
        action='append', metavar='s', choices=avail_tests,
        help='Only include this validator (can use multiple times)')
    parser.add_argument('-x', '--exclude', dest='exclude_tests',
        action='append', metavar='s', choices=avail_tests,
        help='Exclude this validator (can use multiple times)')
    parser.add_argument('-d', '--define', dest='defines',
        action='append', metavar='s',
        help='Define other params (can use multiple times)')
    parser.add_argument('-f', '--force-dnssec-type', dest='force_dnssec_type',
        default='detect', metavar='s',
        choices=['detect', 'unsigned', 'NSEC', 'NSEC3'],
        help='Use DNSSEC type (one of: %(choices)s) [%(default)s]')
    parser.add_argument('-e', '--errors-only', dest='errors_only',
        default=False, action='store_true',
        help='Show validation errors only [%(default)s]')
    parser.add_argument('-v', '--verbose', dest='verbose',
        default=False, action='store_true',
        help='Show detailed processing info [%(default)s]')
    args = parser.parse_args()

    # Fix-up args:
    if args.include_tests is None:
        args.include_tests = []
    if args.exclude_tests is None:
        args.exclude_tests = []
    if args.defines is None:
        args.defines = []
    for sec_param in args.defines:
        (p_name, p_val) = sec_param.split('=')[:2]
        setattr(args, p_name, p_val)

    # Run the app:
    _run_main(avail_loaders, avail_tests, args)


if __name__ == '__main__':
    run()

# end of file



'''
dns_sprockets_impl - DNS Zone validation tool implementation class.
-------------------------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import time

import dns_sprockets_lib.loader_classes as loader_classes
import dns_sprockets_lib.validators as validators
import dns_sprockets_lib.validator_classes as validator_classes


class DNSSprocketsImpl(object):
    # pylint: disable=too-few-public-methods
    '''
    Performs zone validation.
    '''

    def __init__(self, avail_loaders, avail_tests, args):
        '''
        Constructor.

        :param list avail_loaders: List of loader module names.
        :param list avail_tests: List of validator module names.
        :param obj args: The application's command-line args.
        '''
        self.avail_loaders = avail_loaders
        self.avail_tests = avail_tests
        self.args = args

    def run(self):
        '''
        Runs the instance, which loads a zone with the specified loader, and
        loads and runs the specified tests.

        :return: A 3-tuple (loader_elapsed_time, test_count, error_count)
        '''

        # Create the zone loader instance:
        loader_classes.load_all()
        loader_class = loader_classes.ALL_CLASSES[self.args.loader]
        loader = loader_class(self.args)

        # Create the test instances:
        validator_classes.load_all()
        tests_pool = self.args.include_tests or self.avail_tests
        test_names = [t for t in tests_pool if t not in self.args.exclude_tests]
        tests = {
            validators.ZONE_TEST: [],
            validators.NODE_TEST: [],
            validators.RRSET_TEST: [],
            validators.REC_TEST: []}
        for test_name in test_names:
            test_class = validator_classes.ALL_CLASSES[test_name]
            test = test_class(self.args)
            tests[test.TEST_TYPE].append(test)

        # Load the zone:
        load_start_time = time.time()
        zone_obj = loader.run()
        load_time = time.time() - load_start_time
        if not zone_obj:
            raise RuntimeError('No zone to validate!')
        context = validators.Context(self.args, zone_obj)

        # Print some stats:
        print('# Checking zone:', context.zone_name)
        print('# Loader: %s from: %s elapsed=%f secs' % (
            self.args.loader, self.args.source, load_time))
        if self.args.force_dnssec_type == 'detect':
            print('# Zone appears to be DNSSEC type:', context.dnssec_type)
        else:
            print('# Forcing DNSSEC type for zone to:', context.dnssec_type)
        print('# Extra defines:', self.args.defines)

        # Filter tests based on zone's DNSSEC type:
        running_test_names = []
        for tests_type in list(tests):
            validators.dnssec_filter_tests_by_context(tests[tests_type], context)
            running_test_names += [test.TEST_NAME for test in tests[tests_type]]
        print('# Running tests:', running_test_names)

        counts = {'tests': 0, 'errors': 0}

        # Run the zone tests:
        for test in tests[validators.ZONE_TEST]:
            DNSSprocketsImpl._run_test(self.args, test, context, counts)

        # Run the node, rrset, and rec tests:
        for name in context.node_names:
            node = context.zone_obj.get_node(name)
            if node:  # Not present for Empty Non-Terminals

                # Node tests:
                for test in tests[validators.NODE_TEST]:
                    test_node = validators.filter_node(node, test.TEST_RRTYPE)
                    DNSSprocketsImpl._run_test(
                        self.args, test, context, counts,
                        name=name, node=test_node)

                # RRSet tests:
                for rdataset in node.rdatasets:
                    for test in tests[validators.RRSET_TEST]:
                        if validators.test_covers_type(test, rdataset.rdtype):
                            DNSSprocketsImpl._run_test(
                                self.args, test, context, counts,
                                name=name, rdataset=rdataset)

                    # Rec tests:
                    for rdata in rdataset:
                        for test in tests[validators.REC_TEST]:
                            if validators.test_covers_type(test, rdata.rdtype):
                                DNSSprocketsImpl._run_test(
                                    self.args, test, context, counts,
                                    name=name, ttl=rdataset.ttl, rdata=rdata)

        print('# END RESULT: %d ERRORS in %d tests' % (counts['errors'], counts['tests']))
        if hasattr(self.args, 'verbose') and self.args.verbose:
            print('#  TEST TIMES:')
            for key in list(tests):
                for test in tests[key]:
                    print('#  %f secs for %s (%d runs for %f secs each)' % (
                        test.total_time, 
                        test.TEST_NAME, 
                        test.total_runs,
                        test.total_time / test.total_runs))
        return (load_time, counts['tests'], counts['errors'])

    @staticmethod
    def _run_test(args, test, context, counts, **kwargs):
        '''
        Runs an individual test.

        :param obj args: The application's command-line args.
        :param obj test: The validator instance to use.
        :param obj context: The testing context.
        :param dict counts: Dictionary of test and error counts.
        :param dict kwargs: Other parameters specific to the TEST_TYPE being run.
        '''
        suggested_tested = validators.make_suggested_tested(test, context, **kwargs)
        test.start_timer()
        (tested, result) = test.run(context, suggested_tested, **kwargs)
        test.stop_timer()
        if tested is not None:
            test.total_runs += 1
            counts['tests'] += 1
            good = not result
            if not good:
                counts['errors'] += 1
            if not good or not args.errors_only:
                print('TEST %s(%s) => %s' % (
                    test.TEST_NAME,
                    tested,
                    good and 'OK' or 'FAIL: %s' % (result)))

# end of file



'''
nsec3_orphan - Record test: Nsec3Orphan

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import dns.rdatatype

import dns_sprockets_lib.dnssec_nsecx as nsecx
import dns_sprockets_lib.validators as validators


class Nsec3Orphan(validators.RecTest):
    # pylint: disable=too-few-public-methods
    '''
    Checks for orphan or invalid-covers NSEC3s.
    '''
    TEST_DNSSECTYPE = 'NSEC3'
    TEST_RRTYPE = 'NSEC3'

    def __init__(self, args):
        '''
        Ctor, initializes NSEC3-hashed names cache of all node names in zone.
        '''
        super(Nsec3Orphan, self).__init__(args)
        self.hashed_node_names = None

    def run(self, context, suggested_tested, name, ttl, rdata):
        # pylint: disable=too-many-arguments

        tested = None
        result = None

        # Only run test if there's an NSEC3PARAM:
        nsec3param = (len(context.nsec3param_rdataset.items) and
                      context.nsec3param_rdataset.items[0] or None)
        if nsec3param:
            tested = suggested_tested

            # Check and build hashed node names cache if necessary:
            if self.hashed_node_names is None:
                self.hashed_node_names = {}
                if self.args.verbose:
                    print('#  DICTIONARY OF HASHED NODE NAMES TO NODE NAMES:')
                for node_name in context.node_names:
                    hashed_node_name = nsecx.hash_nsec3_name(
                        node_name,
                        nsec3param.salt,
                        nsec3param.algorithm,
                        nsec3param.iterations)
                    self.hashed_node_names[hashed_node_name] = node_name
                    if self.args.verbose:
                        print('#  %s -> %s' % (hashed_node_name.upper(), node_name))

            # Get the NSEC3's lowercased hashed label from owner name:
            hashed_label = name.labels[0].lower().decode()

            # Insure the NSEC3 covers a node name in the zone:
            if hashed_label in self.hashed_node_names:
                node_name = self.hashed_node_names[hashed_label]

                # Get the types covered by those in the node:
                node = context.zone_obj.get_node(node_name)
                node_rdatasets = node and node.rdatasets or []
                node_types = [
                    rdataset.rdtype for rdataset in node_rdatasets
                    if rdataset.rdtype != dns.rdatatype.NSEC3]

                # Get the types covered by the NSEC3:
                covered_types = nsecx.get_covered_types(rdata)

                # Check for any node types not covered:
                not_covered = [t for t in node_types
                               if t not in covered_types]
                if not_covered:
                    result = 'Doesn\'t cover types: %s' % (
                        ' '.join([dns.rdatatype.to_text(t)
                                  for t in not_covered]))

                # Check for any extra covered types not needed:
                extra_covered = [t for t in covered_types
                                 if t not in node_types]
                if extra_covered:
                    if not result:
                        result = ''
                    else:
                        result += ' and '
                    result += 'Covers non-existent RRSet types: %s' % (
                        ' '.join([dns.rdatatype.to_text(t)
                                  for t in extra_covered]))
            else:
                result = 'Doesn\'t cover any RRSet in the zone'

        return (tested, result)

# end of file



'''
test_nsec3_chain - Tests nsec3_chain validator.

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''
# pylint: skip-file


import dns_sprockets_lib.validators.tests.harness as harness


def test_nsec_chain():

    (tests, errs) = harness.test_validator(
        test_name='nsec3_chain',
        zone_name='001.cst.net',
        file_name='001.cst.net.')
    assert tests == 1
    assert errs == 0

    (tests, errs) = harness.test_validator(
        test_name='nsec3_chain',
        zone_name='example',
        file_name='rfc5155_example.')
    assert tests == 1
    assert errs == 0

    (tests, errs) = harness.test_validator(
        test_name='nsec3_chain',
        zone_name='example',
        file_name='rfc5155_example._no_nsec3')
    assert tests == 1
    assert errs == 1

    (tests, errs) = harness.test_validator(
        test_name='nsec3_chain',
        zone_name='example',
        file_name='rfc5155_example._chain_broke')
    assert tests == 1
    assert errs == 1

    (tests, errs) = harness.test_validator(
        test_name='nsec3_chain',
        zone_name='example',
        file_name='rfc5155_example._chain_short')
    assert tests == 1
    assert errs == 1

    (tests, errs) = harness.test_validator(
        test_name='nsec3_chain',
        zone_name='example',
        file_name='rfc5155_example._chain_misorder')
    assert tests == 1
    assert errs == 1

    (tests, errs) = harness.test_validator(
        test_name='nsec3_chain',
        zone_name='signerendtoendtest.com',
        file_name='signerendtoendtest.com.')
    assert tests == 1
    assert errs == 0

# end of file




'''
test_dns_utils- Tests misc. DNS support routines.

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''
# pylint: skip-file


import dns.name

import dns_sprockets_lib.dns_utils as dns_utils


def test_is_delegated():

    tests = [
        ([], '', False),
        ([], 'test', False),
        (['test'], '', False),
        (['test'], 'test', True),
        (['test'], 'a.test', True),
        (['test'], 'b.a.test', True),
        (['test1', 'test2'], 'test', False),
        (['test1', 'test2'], 'test1.test2', True),
        (['test1', 'test2'], 'test2.test1', True)]

    for test in tests:
        names = [dns.name.from_text(s) for s in test[0]]
        name = dns.name.from_text(test[1])
        assert dns_utils.is_delegated(names, name) is test[2]


def test_dnssec_name_cmp_to_key():

    key_class = dns_utils.dns_name_cmp_to_key()
    a = key_class(dns.name.from_text('a.b.c'))
    b = key_class(dns.name.from_text('b.c'))
    assert (a < b) is False
    assert (a > b) is True
    assert (a == b) is False
    assert (a <= b) is False
    assert (a >= b) is True
    assert (a != b) is True


def test_dnssec_sort_names():

    tests = [
        ([], []),
        ([''], ['']),
        (['a'], ['a']),
        (['a', ''], ['', 'a']),
        (['a', 'b'], ['a', 'b']),
        (['b', 'a'], ['a', 'b']),
        (['a.b', 'b'], ['b', 'a.b']),
        (['a.b', 'c.a.b', 'b'], ['b', 'a.b', 'c.a.b']),
        (['aa.b', 'a.b', 'caa.b'], ['a.b', 'aa.b', 'caa.b'])]

    for test in tests:
        ins = [dns.name.from_text(s) for s in test[0]]
        outs = [dns.name.from_text(s) for s in test[1]]
        assert dns_utils.dnssec_sort_names(ins) == outs


def test_calc_node_names():

    tests = [
        ('', False, [], []),
        ('', False, [''], ['']),
        ('', False, ['a'], ['a']),
        ('', False, ['d.c.b.a'], ['d.c.b.a']),
        ('', False, ['*.a'], ['*.a']),
        ('', False, ['a', '*.a'], ['a', '*.a']),
        ('', False, ['b.a', '*.a'], ['*.a', 'b.a']),
        ('', False, ['c.b.a', '*.a'], ['*.a', 'c.b.a']),
        ('', False, ['d.c.b.a', '*.a'], ['*.a', 'd.c.b.a']),
        ('', False, ['d.c.b.a', '*.*.a'], ['*.*.a', 'd.c.b.a']),
        ('', False, ['d.c.b.a', '*.a', '*.*.a'], ['*.a', '*.*.a', 'd.c.b.a']),
        ('a', True, [], []),
        ('a', True, [''], ['']),
        ('a', True, ['a'], ['a']),
        ('a', True, ['d.c.b.a'], ['a', 'b.a', 'c.b.a', 'd.c.b.a']),
        ('a', True, ['*.a'], ['a', '*.a']),
        ('a', True, ['a', '*.a'], ['a', '*.a']),
        ('a', True, ['b.a', '*.a'], ['a', '*.a', 'b.a']),
        ('a', True, ['c.b.a', '*.a'], ['a', '*.a', 'b.a', 'c.b.a']),
        ('a', True, ['d.c.b.a', '*.a'],
                    ['a', '*.a', 'b.a', 'c.b.a', 'd.c.b.a']),
        ('a', True, ['d.c.b.a', '*.*.a'],
                    ['a', '*.a', '*.*.a', 'b.a', 'c.b.a', 'd.c.b.a']),
        ('a', True, ['d.c.b.a', '*.a', '*.*.a'],
                    ['a', '*.a', '*.*.a', 'b.a', 'c.b.a', 'd.c.b.a']),
        ('example', True,
            ['example', 'a.example', 'ns1.a.example', 'ns2.a.example',
             'ai.example', 'c.example', 'ns1.c.example', 'ns2.c.example',
             'ns1.example', 'ns2.example', '*.w.example', 'x.w.example',
             'x.y.w.example', 'xx.example'],
            ['example', 'a.example', 'ns1.a.example', 'ns2.a.example',
             'ai.example', 'c.example', 'ns1.c.example', 'ns2.c.example',
             'ns1.example', 'ns2.example', 'w.example', '*.w.example',
             'x.w.example', 'y.w.example', 'x.y.w.example', 'xx.example'])]

    for test in tests:
        ents_too = test[1]
        zname = ents_too and dns.name.from_text(test[0]) or None
        ins = [dns.name.from_text(s) for s in test[2]]
        outs = [dns.name.from_text(s) for s in test[3]]
        actual = dns_utils.calc_node_names(ins, ents_too, zname)
        #print zname, ents_too, ins, outs, actual
        assert actual == outs

# end of file



'''
soa_unique - Zone test: SoaUnique

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import dns_sprockets_lib.validators as validators


class SoaUnique(validators.ZoneTest):
    # pylint: disable=too-few-public-methods
    '''
    Checks for a single SOA in the zone.
    '''

    def run(self, context, suggested_tested):

        result = None

        # Insure there's only one SOA in the whole zone:
        soa_cnt = 0
        for (_, rdataset) in context.zone_obj.iterate_rdatasets(rdtype='SOA'):
            soa_cnt += len(rdataset)
        if soa_cnt == 0:
            result = 'No SOA in the zone'
        elif soa_cnt > 1:
            result = 'More than one SOA in the zone (count=%d)' % (soa_cnt)

        return (suggested_tested, result)

# end of file



'''
__init__.py - Validators library for dns_sprockets zone validator.
------------------------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''

import time

import dns.rdatatype
import dns.name
import dns.rdata
import dns.rdtypes
import dns.rdtypes.ANY
import dns.node

import dns_sprockets_lib.utils as utils
import dns_sprockets_lib.dns_utils as dns_utils


# Test types:
(ZONE_TEST,   # A test type that validates entire zone.
 NODE_TEST,   # A test type that validates nodes in the zone.
 RRSET_TEST,  # A test type that validates RRSets in the zone.
 REC_TEST     # A test type that validates individual records in the zone.
) = list(range(4))


def test_type_to_str(test_type, test_rrtype=None):
    '''
    Convert a test_type and test_rrtype to a string for output purposes.

    :param int test_type: The TEST_TYPE attribute from the test.
    :param str test_rrtype: The string describing record type(s) covered by the test.
    :return: Description string for test.
    '''
    specific = test_rrtype and '[%s]' % (test_rrtype) or ''
    if test_type == ZONE_TEST:
        return 'ZONE_TEST' + specific
    elif test_type == NODE_TEST:
        return 'NODE_TEST' + specific
    elif test_type == RRSET_TEST:
        return 'RRSET_TEST' + specific
    elif test_type == REC_TEST:
        return 'REC_TEST' + specific


def rec_to_abbrev_text(name, ttl, klass, rdata):
    '''
    Translates a record to abbreviated text.  For most records, this is the
    same as the to_text(); for others (such as RRSIG), it is truncated to
    attempt to fit on a single terminal line.

    :param str name: The owner name of the record.
    :param int ttl: The TTL for the record.
    :param int/str klass: The class of the record.
    :param obj rdata: The rdata of the record.
    :return: Text description of the record.
    '''
    if isinstance(rdata, dns.rdtypes.ANY.RRSIG.RRSIG):
        # pylint: disable=protected-access
        sig = dns.rdata._base64ify(rdata.signature)
        rdata_txt = '%s %d %d %d %s %s %d %s %s' % (
            dns.rdatatype.to_text(rdata.type_covered),
            rdata.algorithm,
            rdata.labels,
            rdata.original_ttl,
            dns.rdtypes.ANY.RRSIG.posixtime_to_sigtime(rdata.expiration),
            dns.rdtypes.ANY.RRSIG.posixtime_to_sigtime(rdata.inception),
            rdata.key_tag,
            rdata.signer,
            '%s...%s' % (sig[:6], sig[len(sig) - 6:]))
    else:
        rdata_txt = rdata.to_text(relativize=False)

    klass_txt = (isinstance(klass, str)
        and klass or dns.rdataclass.to_text(klass))
    return '%s %d %s %s %s' % (
        name, ttl, klass_txt, rdata.__class__.__name__, rdata_txt)


def dnssec_filter_tests_by_context(tests, context):
    '''
    Removes any tests from the tests list that do not apply to the context.

    :param list tests: List of tests to filter.
    :param obj context: The context being used.
    '''
    remove_tests = []

    for test in tests:
        rem_test = None
        if context.dnssec_type == 'unsigned':
            if test.TEST_DNSSECTYPE is not None:
                rem_test = test
        elif context.dnssec_type == 'NSEC':
            if test.TEST_DNSSECTYPE == 'NSEC3':
                rem_test = test
        elif context.dnssec_type == 'NSEC3':
            if test.TEST_DNSSECTYPE == 'NSEC':
                rem_test = test

        if rem_test:
            remove_tests.append(rem_test)
            print('# Skipping test: %s  (DNSSEC type for zone: %s, for test: %s)' % (
                rem_test.TEST_NAME,
                context.dnssec_type,
                test.TEST_DNSSECTYPE))

    for test in remove_tests:
        tests.remove(test)


def filter_node(node, test_rrtype):
    '''
    Returns a node that has rdatasets that match the test RR types.  If the
    test_rrtype is specified, a new, temporary node for use by the validator
    will be generated, which only has those rdatasets mentioned.

    :param obj node: The dns.node.Node to inspect.
    :param str test_rrtype: The string description of RR type(s) that the test covers.
    :return: The dns.node.Node for the validator to examine.
    '''
    if not test_rrtype:
        return node
    rrtypes_txt = test_rrtype.split(',')
    rrtypes = [dns.rdatatype.from_text(t) for t in rrtypes_txt]
    new_node = dns.node.Node()
    new_node.rdatasets = [rds for rds in node.rdatasets
                          if rds.rdtype in rrtypes]
    return new_node


def test_covers_type(test, rdtype):
    '''
    Checks to see if a test covers a RR type.

    :param obj test: The test to examine.
    :param int rdtype: The dns.rdatatype for the rdataset/record under consideration.
    :return: True if the test covers the type; False if not.
    '''
    if not test.TEST_RRTYPE:
        return True
    rrtypes_txt = test.TEST_RRTYPE.split(',')
    rrtypes = [dns.rdatatype.from_text(t) for t in rrtypes_txt]
    return rdtype in rrtypes


def make_suggested_tested(test, context, **kwargs):
    '''
    Generates a description for the test being run.  A test description is
    printed for each test instance being run against zone, node, rrset, or
    record, and this is the suggested description.  Usually, specific test
    instances will use this value for 'tested' return variable, but are free
    to ignore this description in favor of their own if desired.

    :param obj test: The test being run.
    :param obj context: The testing context.
    :param dict kwargs: Optional, test-type-specific parameters.
    :return: A string describing the test instance being run.
    '''
    if test.TEST_TYPE == ZONE_TEST:
        suggested_tested = 'ZONE(%s %s)' % (
            context.zone_name,
            dns.rdataclass.to_text(context.zone_obj.rdclass))
    elif test.TEST_TYPE == NODE_TEST:
        suggested_tested = 'NODE(%s %s)' % (
            kwargs['name'],
            dns.rdataclass.to_text(context.zone_obj.rdclass))
    elif test.TEST_TYPE == RRSET_TEST:
        suggested_tested = 'RRSET(%s %s %s)' % (
            kwargs['name'],
            dns.rdataclass.to_text(context.zone_obj.rdclass),
            dns.rdatatype.to_text(kwargs['rdataset'].rdtype))
    elif test.TEST_TYPE == REC_TEST:
        suggested_tested = 'REC(%s)' % (
            rec_to_abbrev_text(
                kwargs['name'],
                kwargs['ttl'],
                context.zone_obj.rdclass,
                kwargs['rdata']))
    return suggested_tested


class Context(object):
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    '''
    A testing context containing the zone name, zone_obj, etc.
    '''
    def __init__(self, args, zone_obj):
        '''
        Ctor.

        :param obj args: The application arguments.
        :param obj zone_obj: The dns.zone.Zone instance.
        '''
        self.zone_name = dns.name.from_text(args.zone)
        self.zone_obj = zone_obj

        # Get SOA if available:
        self.soa_rdataset = self.zone_obj.get_rdataset(
            self.zone_name, dns.rdatatype.SOA)

        # Get DNSKEY(s) if available:
        self.dnskey_rdataset = self.zone_obj.get_rdataset(
            self.zone_name, dns.rdatatype.DNSKEY)

        # Get NSEC3PARAM(s) if available:
        self.nsec3param_rdataset = self.zone_obj.get_rdataset(
            self.zone_name, dns.rdatatype.NSEC3PARAM)

        # Get delegated zones if any:
        self.delegated_names = [
            name for (name, _) in self.zone_obj.iterate_rdatasets('NS')
            if name != self.zone_name]

        # Force or detect zone's DNSSEC type:
        if args.force_dnssec_type != 'detect':
            self.dnssec_type = args.force_dnssec_type
        else:
            # See if there are any NSEC or NSEC3's:
            has_nsec = next(self.zone_obj.iterate_rdatasets(dns.rdatatype.NSEC), None)
            has_nsec3 = (self.nsec3param_rdataset or
                next(self.zone_obj.iterate_rdatasets(dns.rdatatype.NSEC3), None))

            # See if this appears to be a signed zone (note: can't seem to
            # practically check all RRSIG's since they "cover" other records,
            # which would require us to iterate all possible "covers" values,
            # so just try a few obvious ones):
            seems_signed = (
                self.dnskey_rdataset or
                has_nsec or
                has_nsec3 or
                next(self.zone_obj.iterate_rdatasets(
                    dns.rdatatype.DS), None) or
                next(self.zone_obj.iterate_rdatasets(
                    dns.rdatatype.RRSIG, dns.rdatatype.SOA), None) or
                next(self.zone_obj.iterate_rdatasets(
                    dns.rdatatype.RRSIG, dns.rdatatype.NS), None) or
                next(self.zone_obj.iterate_rdatasets(
                    dns.rdatatype.RRSIG, dns.rdatatype.A), None) or
                next(self.zone_obj.iterate_rdatasets(
                    dns.rdatatype.RRSIG, dns.rdatatype.AAAA), None))

            self.dnssec_type = (
                has_nsec3 and 'NSEC3' or
                has_nsec and 'NSEC' or
                seems_signed and 'NSEC3' or  # assume NSEC3-type
                'unsigned')

        # Get DNSSEC-ordered list of names in zone (including any Empty Non-
        # Terminals if NSEC3-style zone):
        self.node_names = dns_utils.calc_node_names(
            list(zone_obj.nodes),
            self.dnssec_type == 'NSEC3', self.zone_name)

    def is_delegated(self, name):
        '''
        :return: True if name is delegated w.r.t. the context.
        '''
        return dns_utils.is_delegated(self.delegated_names, name)

#
# For validator classes, use short docstrings, which will be used for the actual
# test descriptions!
#


class _Validator(object):
    # pylint: disable=too-few-public-methods
    '''
    [Base class for validator classes]
    '''
    TEST_NAME = None  # Automatically set in __init__.
    TEST_TYPE = None  # Override expected!  e.g.: ZONE_TEST
    TEST_DNSSECTYPE = None  # Override possible!  one of: None, True, 'NSEC' or 'NSEC3'
    TEST_RRTYPE = None  # Override possible!  e.g.: 'A', or 'RRSIG,NSEC3PARAM'
    TEST_OPTARGS = {}  # Override possible!  e.g.: {'now': (None, 'Time to use for now')}

    def __init__(self, args):
        '''
        Ctor, caches the arguments used to run the application, and grabs any
        optional test arguments.
        '''
        # pylint: disable=invalid-name
        self.TEST_NAME = utils.camelcase_to_underscores(self.__class__.__name__)
        self.args = args

        utils.process_optargs(self.TEST_OPTARGS, self.TEST_NAME, self)

        # For accumulating run times:
        self.total_time = 0
        self.start_time = 0
        
        # For accumulating number of runs:
        self.total_runs = 0

    def start_timer(self):
        '''
        Starts the validator timer.
        '''
        self.start_time = time.time()

    def stop_timer(self):
        '''
        Stops the validator timer and adds to accumulator.
        '''
        self.total_time += time.time() - self.start_time


class ZoneTest(_Validator):
    # pylint: disable=too-few-public-methods
    '''
    [Base class for zone-type validators]
    '''
    TEST_TYPE = ZONE_TEST

    def run(self, suggested_tested, context):
        # pylint: disable=unused-argument
        '''
        Runs the zone-type validator.

        :param str suggested_tested: A suggested tested value.
        :param obj context: The testing context.
        :return: A 2-tuple (tested, result)
        '''
        return ('OOPS!', 'ERROR: run() not overridden for %s' % (self.TEST_NAME))


class NodeTest(_Validator):
    # pylint: disable=too-few-public-methods
    '''
    [Base class for node-type validators.  Derived classes *may* be restricted
    to specific RRType's by specifying a TEST_RRTYPE]
    '''
    TEST_TYPE = NODE_TEST

    def run(self, context, suggested_tested, name, node):
        # pylint: disable=unused-argument
        '''
        Runs the node-type validator.  If a TEST_RRTYPE specified, the node
        presented to the validator will be filtered accordingly.

        :param obj context: The testing context.
        :param str suggested_tested: A suggested tested value.
        :param str name: The name being tested.
        :param obj node: The dns.Node corresponding to the name.
        :return: A 2-tuple (tested, result)
        '''
        return ('OOPS!', 'ERROR: run() not overridden for %s' % (self.TEST_NAME))


class RRSetTest(_Validator):
    # pylint: disable=too-few-public-methods
    '''
    [Base class for rrset-type validators.  Derived classes *may* be restricted
    to specific RRType's by specifying a TEST_RRTYPE]
    '''
    TEST_TYPE = RRSET_TEST

    def run(self, context, suggested_tested, name, rdataset):
        # pylint: disable=unused-argument
        '''
        Runs the name-type validator.  If a TEST_RRTYPE is specified, the RRSet
        presented to the validator will be filtered accordingly.

        :param obj context: The testing context.
        :param str suggested_tested: A suggested tested value.
        :param str name: The name being tested.
        :param obj rdataset: The dns.rdataset corresponding to the name.
        :return: A 2-tuple (tested, result)
        '''
        return ('OOPS!', 'ERROR: run() not overridden for %s' % (self.TEST_NAME))


class RecTest(_Validator):
    # pylint: disable=too-few-public-methods
    '''
    [Base class for record-type validators.  Derived classes *may* be restricted
    to specific RRType's by specifying a TEST_RRTYPE]
    '''
    TEST_TYPE = REC_TEST

    def run(self, context, suggested_tested, name, ttl, rdata):
        # pylint: disable=too-many-arguments
        # pylint: disable=unused-argument
        '''
        Runs the record-type validator.  If a TEST_RRTYPE is specified, the
        validator will only see those types of records.

        :param obj context: The testing context.
        :param str suggested_tested: A suggested tested value.
        :param str name: The name of the record being tested.
        :param int ttl: The TTL of the record being tested.
        :param obj rdata: The dns.rdata.Rdata object being tested.
        :return: A 2-tuple (tested, result)
        '''
        return ('OOPS!', 'ERROR: run() not overridden for %s' % (self.TEST_NAME))

# end of file



'''
loader_classes - Functions for dealing with zone loaders
--------------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import importlib

import dns_sprockets_lib.utils as utils
import dns_sprockets_lib.loaders as loaders


# Dictionary of loader_name -> loader_class:
ALL_CLASSES = {}


def load_all():
    '''
    Scans the sprocket_support.loaders package for "public" modules and loads
    their classes into the ALL_CLASSES dictionary.
    '''
    if not ALL_CLASSES:
        avail_loaders = utils.public_modules_in_package(loaders, ['tests'])
        for loader_name in avail_loaders:
            loader_module_name = 'dns_sprockets_lib.loaders.%s' % (loader_name)
            loader_module = importlib.import_module(loader_module_name)
            ALL_CLASSES[loader_name] = getattr(
                loader_module, utils.underscores_to_camelcase(loader_name))


def _get_descriptions():
    '''
    Get a list of descriptions of all loaders in the ALL_CLASSES dictionary.
    Each description is a tuple: (loader_name, doc_string, loader_optargs).
    The returned list is ordered by loader_name.
    '''
    load_all()
    return sorted(
        [(key, val.__doc__.strip(), val.LOADER_OPTARGS)
         for (key, val) in ALL_CLASSES.items()])


def get_formatted_descriptions():
    '''
    Gets a formatted string containing descriptions of all loaders in the
    ALL_CLASSES dictionary.
    '''
    descs = _get_descriptions()
    return '\n'.join(
        ['LOADER: %s - %s%s' % (
            ld[0],
            ld[1],
            ''.join(sorted(
                ['\n    DEFINE: (%s_)%s - %s (default=%s)' % (
                    ld[0], opt_name, opt_val[1], opt_val[0])
                 for (opt_name, opt_val) in ld[2].items()])))
         for ld in descs])

# end of file



'''
test_file - Tests dns_sprockets_lib.loaders.file module.

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''
# pylint: skip-file


import dns.rdatatype
import dns.zone

from dns_sprockets_lib.loaders.file import File


def test_File():

    class test_args(object):
        source = 'dns_sprockets_lib/tests/data/001.cst.net.'
        zone = '001.cst.net'

    test_file = File(test_args())
    assert test_file.rdclass == 'IN'
    assert test_file.allow_include == '1'

    zone_obj = test_file.run()
    assert isinstance(zone_obj, dns.zone.Zone)
    assert zone_obj.get_rdataset('001.cst.net.', dns.rdatatype.SOA)

# end of file



'''
rrsig_time - Record test: RrsigTime

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import dns_sprockets_lib.validators as validators


class RrsigTime(validators.RecTest):
    # pylint: disable=too-few-public-methods
    '''
    Checks RRSIG's inception <= expiration.
    '''
    TEST_DNSSECTYPE = True
    TEST_RRTYPE = 'RRSIG'

    def run(self, context, suggested_tested, name, ttl, rdata):
        # pylint: disable=too-many-arguments

        result = None
        if rdata.inception > rdata.expiration:
            result = 'Inception time greater than expiration time'
        return (suggested_tested, result)

# end of file



'''
rrsig_ttls_match - Record test: RrsigTtlsMatch

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import dns_sprockets_lib.validators as validators


class RrsigTtlsMatch(validators.RecTest):
    # pylint: disable=too-few-public-methods
    '''
    Checks RRSIG TTL's match original and covered TTL's.
    '''
    TEST_DNSSECTYPE = True
    TEST_RRTYPE = 'RRSIG'

    def run(self, context, suggested_tested, name, ttl, rdata):
        # pylint: disable=too-many-arguments

        result = None

        if ttl != rdata.original_ttl:
            result = 'TTL doesn\'t match original TTL'

        rdataset = context.zone_obj.get_rdataset(name, rdata.type_covered)
        if rdataset and ttl != rdataset.ttl:
            if result:
                result += ', and '
            else:
                result = ''
            result += 'TTL doesn\'t match covered TTL'

        return (suggested_tested, result)

# end of file



'''
validator_classes - Functions for dealing with validators.
----------------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import importlib

import dns_sprockets_lib.utils as utils
import dns_sprockets_lib.validators as validators


# Dictionary of validator_name -> validator_class:
ALL_CLASSES = {}


def load_all():
    '''
    Scans the sprocket_support.validators package for "public" modules and loads
    their classes into the ALL_CLASSES dictionary.
    '''
    if not ALL_CLASSES:
        avail_tests = utils.public_modules_in_package(validators, ['tests'])
        for test_name in avail_tests:
            test_module_name = 'dns_sprockets_lib.validators.%s' % (test_name)
            test_module = importlib.import_module(test_module_name)
            ALL_CLASSES[test_name] = getattr(
                test_module, utils.underscores_to_camelcase(test_name))


def _get_descriptions():
    '''
    Get a list of descriptions of all validators in the ALL_CLASSES dictionary.
    Each description is a tuple: (validator_name, doc_string, test_type,
    test_rrtype, validator_optargs).  The returned list is ordered by
    validator_name.
    '''
    load_all()
    return sorted(
        [(key, val.__doc__.strip(), val.TEST_TYPE, val.TEST_RRTYPE, val.TEST_OPTARGS)
         for (key, val) in ALL_CLASSES.items()])


def get_formatted_descriptions():
    '''
    Gets a formatted string containing descriptions of all validators in the
    ALL_CLASSES dictionary.
    '''
    descs = _get_descriptions()
    return '\n'.join(
        ['TEST: %s (%s) - %s%s' % (
            td[0],
            validators.test_type_to_str(td[2], td[3]),
            td[1],
            ''.join(sorted(
                ['\n    DEFINE: (%s_)%s - %s (default=%s)' % (
                    td[0], opt_name, opt_val[1], opt_val[0])
                 for (opt_name, opt_val) in td[4].items()])))
         for td in descs])

# end of file



'''
dns_sprockets_version - Contains version for dns_sprockets.
-----------------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''

VERSION = '2.0.0'

# end of file



'''
test___init__ - Tests dns_sprockets_lib.validators package.

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''
# pylint: skip-file


import dns.rdatatype

from dns_sprockets_lib.loaders.file import File
import dns_sprockets_lib.validators as validators


def test_test_type_to_str():

    tests = {
        ((validators.ZONE_TEST, None), 'ZONE_TEST'),
        ((validators.ZONE_TEST, ''), 'ZONE_TEST'),
        ((validators.ZONE_TEST, 'SOA'), 'ZONE_TEST[SOA]'),
        ((validators.ZONE_TEST, 'SOA,NS'), 'ZONE_TEST[SOA,NS]'),
        ((validators.NODE_TEST, 'A,AAAA'), 'NODE_TEST[A,AAAA]'),
        ((validators.RRSET_TEST, None), 'RRSET_TEST'),
        ((validators.REC_TEST, None), 'REC_TEST')}

    for test in tests:
        assert validators.test_type_to_str(test[0][0], test[0][1]) == test[1]


def test_dnssec_filter_tests_by_context():

    class test1(object):
        TEST_NAME = 'test1'
        TEST_DNSSECTYPE = None

    class test2(object):
        TEST_NAME = 'test2'
        TEST_DNSSECTYPE = True

    class test3(object):
        TEST_NAME = 'test3'
        TEST_DNSSECTYPE = 'NSEC'

    class test4(object):
        TEST_NAME = 'test4'
        TEST_DNSSECTYPE = 'NSEC3'

    class context1(object):
        dnssec_type = 'unsigned'

    class context2(object):
        dnssec_type = 'NSEC'

    class context3(object):
        dnssec_type = 'NSEC3'

    tests = {
        ((test1(), context1()), 1),
        ((test1(), context2()), 1),
        ((test1(), context3()), 1),
        ((test2(), context1()), 0),
        ((test2(), context2()), 1),
        ((test2(), context3()), 1),
        ((test3(), context1()), 0),
        ((test3(), context2()), 1),
        ((test3(), context3()), 0),
        ((test4(), context1()), 0),
        ((test4(), context2()), 0),
        ((test4(), context3()), 1)}

    for test in tests:
        valis = [test[0][0]]
        context = test[0][1]
        count = test[1]
        validators.dnssec_filter_tests_by_context(valis, context)
        assert len(valis) == count


def test_filter_node():

    class test_args(object):
        source = 'dns_sprockets_lib/tests/data/001.cst.net.'
        zone = '001.cst.net'

    test_file = File(test_args())
    zone_obj = test_file.run()
    node = zone_obj.get_node('001.cst.net.')

    tests = {
        (None, 8),
        ('', 8),
        ('SOA', 1),
        ('A', 0),
        ('SOA,NS', 2),
        ('SOA,NS,DNSKEY', 3),
        ('SOA,NS,DNSKEY,A', 3)}

    for test in tests:
        assert len(validators.filter_node(node, test[0]).rdatasets) == test[1]


def test_test_covers_type():

    class TestInst(object):
        TEST_RRTYPE = None

    tests = {
        ((None, 'SOA'), True),
        (('', 'SOA'), True),
        (('SOA', 'SOA'), True),
        (('SOA', 'A'), False),
        (('SOA,NS', 'A'), False),
        (('SOA,NS', 'SOA'), True),
        (('SOA,NS', 'NS'), True)}

    for test in tests:
        test_inst = TestInst()
        test_inst.TEST_RRTYPE = test[0][0]
        rec_type = dns.rdatatype.from_text(test[0][1])
        assert validators.test_covers_type(test_inst, rec_type) is test[1]


def test_Context():

    class test_args(object):
        source = 'dns_sprockets_lib/tests/data/001.cst.net.'
        zone = '001.cst.net'
        force_dnssec_type = 'detect'

    test_args_inst = test_args()
    test_file = File(test_args_inst)
    zone_obj = test_file.run()
    context = validators.Context(test_args_inst, zone_obj)

    node_names = [
        dns.name.from_text('001.cst.net.'),
        dns.name.from_text('001.001.cst.net.'),
        dns.name.from_text('06ES9CGGDRORFDD4NS9AHOCAIKLDRRP8.001.cst.net.'),
        dns.name.from_text('AL016ABKH6LVDIG6503FS92KDMOTQH4V.001.cst.net.'),
        dns.name.from_text('KQGPU8I0AI43NEM212BD0079J5SI5R3K.001.cst.net.'),
        dns.name.from_text('test.001.cst.net.'),
        dns.name.from_text('test2.001.cst.net.'),
        dns.name.from_text('UQML1AM96TFTFMLKAGTBS82ISR050SH0.001.cst.net.')]

    assert context.zone_name == dns.name.from_text(test_args_inst.zone)
    assert context.zone_obj == zone_obj
    assert context.node_names == node_names
    assert len(context.soa_rdataset.items) == 1
    assert len(context.dnskey_rdataset) == 2
    assert len(context.nsec3param_rdataset) == 1
    assert context.delegated_names == []
    assert context.dnssec_type == 'NSEC3'
    assert context.is_delegated('test.001.cst.net.') is False


def test__Validator():

    class test_args(object):
        test_validator_verbose = True
        test_validator_now = '20150101000000'

    class TestValidator(validators._Validator):
        TEST_OPTARGS = {'now': (None, 'Time to use for now')}

    val = TestValidator(test_args())
    assert val.TEST_TYPE is None
    assert val.TEST_DNSSECTYPE is None
    assert val.TEST_RRTYPE is None
    assert val.TEST_NAME == 'test_validator'
    assert not hasattr(val, 'verbose')
    assert val.now == test_args().test_validator_now


def test_ZoneTest():

    class TestZoneTest(validators.ZoneTest):
        TEST_OPTARGS = {'now': ('20150101000000', 'Time to use for now')}

    class test_args(object):
        verbose = True
        test_zone_test_now = '20151231000000'

    test = TestZoneTest(test_args())
    assert test.TEST_TYPE == validators.ZONE_TEST
    assert test.TEST_NAME == 'test_zone_test'
    assert not hasattr(test, 'verbose')
    assert test.now == test_args.test_zone_test_now


def test_NodeTest():

    class TestNodeTest(validators.NodeTest):
        TEST_OPTARGS = {'now': ('20150101000000', 'Time to use for now')}

    class test_args(object):
        verbose = True
        test_node_test_now = '20151231000000'

    test = TestNodeTest(test_args())
    assert test.TEST_TYPE == validators.NODE_TEST
    assert test.TEST_NAME == 'test_node_test'
    assert not hasattr(test, 'verbose')
    assert test.now == test_args.test_node_test_now


def test_RRSetTest():

    class TestRrsetTest(validators.RRSetTest):
        TEST_OPTARGS = {'now': ('20150101000000', 'Time to use for now')}

    class test_args(object):
        verbose = True
        test_rrset_test_now = '20151231000000'

    test = TestRrsetTest(test_args())
    assert test.TEST_TYPE == validators.RRSET_TEST
    assert test.TEST_NAME == 'test_rrset_test'
    assert not hasattr(test, 'verbose')
    assert test.now == test_args.test_rrset_test_now


def test_RecTest():

    class TestRecTest(validators.RecTest):
        TEST_OPTARGS = {'now': ('20150101000000', 'Time to use for now')}

    class test_args(object):
        verbose = True
        test_rec_test_now = '20151231000000'

    test = TestRecTest(test_args())
    assert test.TEST_TYPE == validators.REC_TEST
    assert test.TEST_NAME == 'test_rec_test'
    assert not hasattr(test, 'verbose')
    assert test.now == test_args.test_rec_test_now

# end of file



'''
ValidatorTestHarness - Test harness for validators.
---------------------------------------------------

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import dns_sprockets_lib.utils as utils
import dns_sprockets_lib.loaders as loaders
import dns_sprockets_lib.validators as validators
from dns_sprockets_lib.dns_sprockets_impl import DNSSprocketsImpl


def test_validator(test_name, zone_name, file_name,
                   extra_defines=None, dnssec_type='detect'):
    '''
    Tests a validator using SprocketImpl instance.

    :param str test_name: The name of the test to run.
    :param str zone_name: The name of the zone to load.
    :param str file_name: The name of the file to load (w.r.t. dns_sprockets_lib/tests/data)
    :param list extra_defines: List of extra defines, in command-line format.
    :param str dnssec_type: Set to force dnssec_type.
    :return: A 2-tuple (test_cnt, error_cnt).
    '''

    class TestArgs(object):
        # pylint: disable=too-few-public-methods
        '''
        Simulated command-line arguments.
        '''
        zone = zone_name
        loader = 'file'
        source = 'dns_sprockets_lib/tests/data/%s' % (file_name)
        include_tests = [test_name]
        exclude_tests = []
        force_dnssec_type = dnssec_type
        errors_only = False
        defines = extra_defines
        verbose = False

    avail_loaders = utils.public_modules_in_package(loaders, ['tests'])
    avail_tests = utils.public_modules_in_package(validators, ['tests'])

    test_args_inst = TestArgs()
    if test_args_inst.defines is None:
        test_args_inst.defines = []
    for sec_param in test_args_inst.defines:
        (p_name, p_val) = sec_param.split('=')[:2]
        setattr(test_args_inst, p_name, p_val)

    sprocket = DNSSprocketsImpl(avail_loaders, avail_tests, test_args_inst)
    (_, test_cnt, err_cnt) = sprocket.run()
    return (test_cnt, err_cnt)

# end of file



'''
dnssec_nsec3 - Stuff for dealing with NSEC3-type zones.
-------------------------------------------------------
(maybe contribute somehow to dnspython?)

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import base64
import codecs

import Crypto.Hash.SHA
import dns.name
import dns.rdtypes.ANY.NSEC3


def encode_salt(b_bin_salt):
    '''
    Encodes binary NSEC3-type "salt" values to hexadecimal representation.

    :param bytes b_bin_salt: The binary salt to encode.
    :return: Lowercase hexadecimal bytes string representating the salt (or None for failure).
    '''
    try:
        b_hex_salt = codecs.encode(b_bin_salt, 'hex-codec')
        return codecs.decode(b_hex_salt, 'latin-1')
    except Exception:
        return None


def decode_salt(hex_salt):
    '''
    Decodes hexedecimal representation of NSEC3-type "salt" values to binary.

    :param str hex_salt: The hex-encoded salt to decode.
    :return: Binary value bytes of the salt value (or None for failure).
    '''
    try:
        b_hex_salt = codecs.encode(hex_salt, 'latin-1')
        return codecs.decode(b_hex_salt, 'hex-codec')
    except Exception:
        return None


def hash_nsec3_name(name, salt, algo, addl_iters, salt_is_binary=True):
    '''
    Hashes a domain name using indicated hashing algorithm / iterations.

    :param (str or dns.name.Name) name: The domain name to hash.
    :param (str or bytes) salt: The encoded salt to use (or empty string).
    :param int algo: The hashing algorithm to use.
    :param int addl_iters: The additional iterations to be applied.
    :param bool salt_is_binary: Set True if the salt is binary; False if hex string.
    :return: The (lowercase) base32hex-encoded string result (or None for failure).
    '''

    # Check name:
    if isinstance(name, dns.name.Name):
        name = name.to_text()
    elif not isinstance(name, str):
        return None

    # We only know how to do SHA-1:
    if algo != dns.rdtypes.ANY.NSEC3.SHA1:
        return None

    # Decode salt if necessary:
    if salt_is_binary:
        if isinstance(salt, str):
            b_salt = codecs.encode(salt, 'latin-1')
        elif isinstance(salt, bytes):
            b_salt = salt
        else:
            return None
    else:
        if not isinstance(salt, str):
            return None
        b_salt = decode_salt(salt)

    # Convert name to lowercase and wire format:
    b_bin_hash = dns.name.from_text(name.lower()).to_wire()

    # Iterate hashing according to RFC5155 section 5:
    for _ in range(1 + int(addl_iters)):
        b_bin_hash = Crypto.Hash.SHA.new(b_bin_hash + b_salt).digest()

    # Convert result to lowercase base32hex:
    b_hash = base64.b32encode(b_bin_hash).translate(
        dns.rdtypes.ANY.NSEC3.b32_normal_to_hex).lower()
    return codecs.decode(b_hash, 'latin-1')


def _windows_covers(windows, rrtype):
    '''
    Check to see if an NSECx windows list covers a type.

    :param list windows: An NSEC or NSEC3 windows list.
    :param int rrtype: The dns.rdatatype to test.
    :return: True if the windows covers the type; False if not.
    '''
    window = rrtype // 256
    for win in windows:
        if win[0] == window:
            bitmap = win[1]
            offset = rrtype % 256
            byte = offset // 8
            if bitmap and len(bitmap) >= byte + 1:
                bit = offset % 8
                return bool(bitmap[byte] & (0x80 >> bit))

    return False


def covers(nsecx_rdata, rrtype):
    '''
    Check to see if an NSECx covers a type.

    :param obj nsec_rdata: An NSEC or NSEC3 instance.
    :param int rrtype: The dns.rdatatype to test.
    :return: True if the NSECx covers the type; False if not.
    '''
    return _windows_covers(nsecx_rdata.windows, rrtype)


def _windows_get_covered_types(windows):
    '''
    Gets list of types covered by an NSECx windows list.

    :param list windows: An NSEC or NSEC3 windows list.
    :return: List of dns.rdatatype's that the NSECx covers.
    '''
    types = []
    for win in windows:
        window = win[0]
        bitmap = win[1]
        if bitmap:
            for byte in range(len(bitmap)):
                bits = bitmap[byte]
                if bits:
                    for bit in range(8):
                        if bits & (0x80 >> bit):
                            types.append(window * 256 + byte * 8 + bit)
    return types


def get_covered_types(nsecx_rdata):
    '''
    Gets list of types covered by an NSECx.

    :param obj nsecx_rdata: The NSEC or NSEC3 instance.
    :return: List of dns.rdatatype's that the NSECx covers.
    '''
    return _windows_get_covered_types(nsecx_rdata.windows)

# end of file



'''
rrsig_missing - RRSet test: RrsigMissing

.. Copyright (c) 2015 Neustar, Inc. All rights reserved.
.. See COPYRIGHT.txt for full notice.  See LICENSE.txt for terms and conditions.
'''


import time

import dns.rdatatype
import dns.rdtypes.ANY.RRSIG
import dns.dnssec

import dns_sprockets_lib.validators as validators


class RrsigMissing(validators.RRSetTest):
    # pylint: disable=too-few-public-methods
    '''
    Checks that all (non-RRSIG, non-delegated) RRSets are covered with an RRSIG.
    '''
    TEST_DNSSECTYPE = True
    TEST_OPTARGS = {
        'now': (None, 'Time to use for validating RRSIG time windows, e.g. 20150101123000'),
        'now_offset': (None, 'Number of seconds to offset the "now" value, e.g. -86400)')}

    def __init__(self, args):

        self.now = None
        self.now_offset = None
        super(RrsigMissing, self).__init__(args)

        self.posix_now = (self.now
            and dns.rdtypes.ANY.RRSIG.sigtime_to_posixtime(self.now)
            or int(time.time()))
        if self.now_offset:
            self.posix_now += int(self.now_offset)

    def run(self, context, suggested_tested, name, rdataset):

        tested = None
        result = None

        # Only do this test for non-RRSIG, non-delegated RRSets:
        if (rdataset.rdtype != dns.rdatatype.RRSIG
            and not context.is_delegated(name)):

            tested = suggested_tested

            # Make sure there's a valid RRSIG for the rdataset:
            rrsigset = context.zone_obj.get_rdataset(
                name, 'RRSIG', covers=rdataset.rdtype)
            if not rrsigset:
                result = 'No RRSIG for name: %s' % (name)
            else:
                # Use dnspython's RRSIG validator:
                try:
                    dns.dnssec.validate(
                        (name, rdataset),
                        (name, rrsigset),
                        {context.zone_name: context.dnskey_rdataset},
                        now=self.posix_now)
                except dns.dnssec.UnsupportedAlgorithm as err:
                    result = str(err)
                except dns.dnssec.ValidationFailure as err:
                    result = str(err)

        return (tested, result)

# end of file
